

set -ex



python -c "import numpy; numpy._pytesttester._show_numpy_info()"
python -c "import scipy; scipy.test(verbose=2, label='fast', tests=['scipy.linalg', 'scipy.fft'], extra_argv=['--durations=50'])"
exit 0
